/*
 *  Code extracted from
 *  linux/kernel/hd.c
 *
 *  Copyright (C) 1991-1998  Linus Torvalds
 *
 *
 *  Thanks to Branko Lankester, lankeste@fwi.uva.nl, who found a bug
 *  in the early extended-partition checks and added DM partitions
 *
 *  Support for DiskManager v6.0x added by Mark Lord,
 *  with information provided by OnTrack.  This now works for linux fdisk
 *  and LILO, as well as loadlin and bootln.  Note that disks other than
 *  /dev/hda *must* have a "DOS" type 0x51 partition in the first slot (hda1).
 *
 *  More flexible handling of extended partitions - aeb, 950831
 *
 *  Added needed MAJORS for new pairs, {hdi,hdj}, {hdk,hdl}
 *
 *  Check partition table on IDE disks for common CHS translations
 *
 *  Moved partition table parsers into separate files and modularized them
 *  (10/08/97 Roman Hodek)
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/blk.h>
#include <linux/init.h>
#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif

#include <asm/system.h>

struct gendisk *gendisk_head = NULL;

int current_minor = 0;
extern int *blk_size[];
extern void rd_load(void);
extern void initrd_load(void);

extern int chr_dev_init(void);
extern int blk_dev_init(void);
extern int scsi_dev_init(void);
extern int net_dev_init(void);

struct partition_parser *partition_parsers = NULL;

#ifdef CONFIG_PPC
extern void note_bootable_part(kdev_t dev, int part);
#endif

/*
 * disk_name() is used by genhd.c and md.c.
 * It formats the devicename of the indicated disk
 * into the supplied buffer, and returns a pointer
 * to that same buffer (for convenience).
 */
char *disk_name (struct gendisk *hd, int minor, char *buf)
{
	unsigned int part;
	const char *maj = hd->major_name;
	int unit = (minor >> hd->minor_shift) + 'a';

	/*
	 * IDE devices use multiple major numbers, but the drives
	 * are named as:  {hda,hdb}, {hdc,hdd}, {hde,hdf}, {hdg,hdh}..
	 * This requires special handling here.
	 */
	switch (hd->major) {
		case IDE5_MAJOR:
			unit += 2;
		case IDE4_MAJOR:
			unit += 2;
		case IDE3_MAJOR:
			unit += 2;
		case IDE2_MAJOR:
			unit += 2;
		case IDE1_MAJOR:
			unit += 2;
		case IDE0_MAJOR:
			maj = "hd";
			break;
	}
	part = minor & ((1 << hd->minor_shift) - 1);
	if (hd->major >= SCSI_DISK1_MAJOR && hd->major <= SCSI_DISK7_MAJOR) {
		unit = unit + (hd->major - SCSI_DISK1_MAJOR + 1) * 16;
		if (unit > 'z') {
			unit -= 'z' + 1;
			sprintf(buf, "sd%c%c", 'a' + unit / 26, 'a' + unit % 26);
			if (part)
				sprintf(buf + 4, "%d", part);
			return buf;
		}
	}
	if (part)
		sprintf(buf, "%s%c%d", maj, unit, part);
	else
		sprintf(buf, "%s%c", maj, unit);
	return buf;
}

void add_partition (struct gendisk *hd, int minor, int start, int size)
{
	char buf[8];
	hd->part[minor].start_sect = start;
	hd->part[minor].nr_sects   = size;
	printk(" %s", disk_name(hd, minor, buf));
}

unsigned int get_ptable_blocksize(kdev_t dev)
{
  int ret = 1024;

  /*
   * See whether the low-level driver has given us a minumum blocksize.
   * If so, check to see whether it is larger than the default of 1024.
   */
  if (!blksize_size[MAJOR(dev)])
    {
      return ret;
    }

  /*
   * Check for certain special power of two sizes that we allow.
   * With anything larger than 1024, we must force the blocksize up to
   * the natural blocksize for the device so that we don't have to try
   * and read partial sectors.  Anything smaller should be just fine.
   */
  switch( blksize_size[MAJOR(dev)][MINOR(dev)] )
    {
    case 2048:
      ret = 2048;
      break;
    case 4096:
      ret = 4096;
      break;
    case 8192:
      ret = 8192;
      break;
    case 1024:
    case 512:
    case 256:
    case 0:
      /*
       * These are all OK.
       */
      break;
    default:
      panic("Strange blocksize for partition table\n");
    }

  return ret;

}


#if CONFIG_KERNELD && CONFIG_FOREIGN_PARTITIONS

#include <linux/kerneld.h>

#ifdef CONFIG_AMIGA_PARTITION_MODULE
#include <linux/affs_hardblocks.h>
static int probe_amiga_partition(struct gendisk *hd, kdev_t dev,
				 struct buffer_head *bh_0)
{
	struct buffer_head *bh;
	int blk, res;

	if (*(u32 *)bh_0->b_data == cpu_to_be32(IDNAME_RIGIDDISK))
		return 1;

	/* also need to check other possible RDB blocks */
	for (blk = 1; blk < RDB_ALLOCATION_LIMIT; blk++) {
		if (!(bh = bread(dev,blk,512)))
			return -1;
		res = *(u32 *)bh->b_data == cpu_to_be32(IDNAME_RIGIDDISK);
		brelse(bh);
		if (res)
			return 1;
	}
	return 0;
}
#endif

#ifdef CONFIG_ATARI_PARTITION_MODULE
#include <linux/atari_rootsec.h>
#include <linux/ctype.h>
#define	VALID_PARTITION(pi,hdsiz)					     \
    (((pi)->flg & 1) &&							     \
     isalnum((pi)->id[0]) && isalnum((pi)->id[1]) && isalnum((pi)->id[2]) && \
     be32_to_cpu((pi)->st) <= (hdsiz) &&				     \
     be32_to_cpu((pi)->st) + be32_to_cpu((pi)->siz) <= (hdsiz))
static int probe_atari_partition(struct gendisk *hd, kdev_t dev,
				 struct buffer_head *bh_0)
{
	struct rootsector *rs;
	u32 hd_size;
	int res;

	rs = (struct rootsector *)bh_0->b_data;
	hd_size = hd->part[0].nr_sects;
	res = VALID_PARTITION( &rs->part[0], hd_size ) ||
	      VALID_PARTITION( &rs->part[1], hd_size ) ||
	      VALID_PARTITION( &rs->part[2], hd_size ) ||
	      VALID_PARTITION( &rs->part[3], hd_size );
	return res;
}
#endif

/*
 * This table lists the magics to recognize the partition formats that are
 * availble as modules. The magics can be stored as 16 or 32 bit little-endian
 * or big-endian numbers. For the more complicated cases (Amiga and Atari), a
 * function doing the real work is called.
 */

enum partbl_probe_type { BE16, BE32, LE16, LE32, FUNC };

static struct partbl_probe {
	char *name;
	enum partbl_probe_type type;
	u32 offset;
	u32 magic;
	int (*func)(struct gendisk *hd, kdev_t dev, struct buffer_head *bh_0);
} partbl_modules[] = {
#ifdef CONFIG_MSDOS_PARTITION_MODULE
	{ "msdos", LE16, 510, 0xAA55, NULL },
#endif
#ifdef CONFIG_OSF_PARTITION_MODULE
	{ "osf",   LE32,  64, 0x82564557, NULL },
#endif
#ifdef CONFIG_SUN_PARTITION_MODULE
	{ "sun",   BE16, 508, 0xDABE, NULL },
#endif
#ifdef CONFIG_AMIGA_PARTITION_MODULE
	{ "amiga", FUNC,   0,      0, probe_amiga_partition },
#endif
#ifdef CONFIG_ATARI_PARTITION_MODULE
	{ "atari", FUNC,   0,      0, probe_atari_partition },
#endif
#ifdef CONFIG_MAC_PARTITION_MODULE
	{ "mac",   BE16,   0, 0x4552, NULL },
#endif
};

#define N_PARTBL_MODULES sizeof(partbl_modules)/sizeof(*partbl_modules)

/*
 * Try to load all partition parsers that could apply to this device
 */
static int try_to_load_parser(struct gendisk *hd, kdev_t dev)
{
	int i, hit = 0, loaded = 0;
	char modname[24];
	struct partbl_probe *mod;
	struct partition_parser *p;
	struct buffer_head *bh;

	if (!(bh = bread(dev,0,get_ptable_blocksize(dev))))
		return -1;

	for( i = 0, mod = partbl_modules; i < N_PARTBL_MODULES; ++i, ++mod ) {
		/* don't probe already loaded formats again */
		for( p = partition_parsers; p; p = p->next )
			if (!strcmp( mod->name, p->name))
				goto skip;
		
		switch( mod->type ) {
		    case LE16:
			hit = le16_to_cpu(*(u16 *)(bh->b_data+mod->offset)) ==
			      mod->magic;
			break;
		    case LE32:
			hit = le32_to_cpu(*(u32 *)(bh->b_data+mod->offset)) ==
			      mod->magic;
			break;
		    case BE16:
			hit = be16_to_cpu(*(u16 *)(bh->b_data+mod->offset)) ==
			      mod->magic;
			break;
		    case BE32:
			hit = be32_to_cpu(*(u32 *)(bh->b_data+mod->offset)) ==
			      mod->magic;
			break;
		    case FUNC:
			hit = mod->func(hd, dev, bh) > 0;
			break;
		}
		if (hit) {
			strcpy(modname, "part-format-");
			strcat(modname, mod->name);
			request_module(modname);
			++loaded;
		}
	    skip:
	}
	brelse(bh);
	return loaded;
}

#endif /* CONFIG_KERNELD && CONFIG_FOREIGN_PARTITIONS */


static void check_partition(struct gendisk *hd, kdev_t dev)
{
	static int first_time = 1;
	int tries = 0;
	unsigned long first_sector;
	char buf[8];
	struct partition_parser *parser;

	if (first_time)
		printk("Partition check:\n");
	first_time = 0;
	first_sector = hd->part[MINOR(dev)].start_sect;

	/*
	 * This is a kludge to allow the partition check to be
	 * skipped for specific drives (e.g. IDE CD-ROM drives)
	 */
	if ((int)first_sector == -1) {
		hd->part[MINOR(dev)].start_sect = 0;
		return;
	}

#if CONFIG_KERNELD && CONFIG_FOREIGN_PARTITIONS
    retry:
#endif
	printk(" %s:", disk_name(hd, MINOR(dev), buf));
	for( parser = partition_parsers; parser; parser = parser->next ) {
		if (parser->func(hd, dev, first_sector))
			return;
	}
	++tries;
#if CONFIG_KERNELD && CONFIG_FOREIGN_PARTITIONS
	if (tries == 1 && try_to_load_parser(hd, dev))
		goto retry;
#endif
	printk(" unknown partition table\n");
}

int register_partbl( struct partition_parser *parser )
{
	struct partition_parser **p;
    
	if (!parser)
		return -EINVAL;
	if (parser->next)
		return -EBUSY;
	for( p = &partition_parsers; *p; p = &(*p)->next )
		if (*p == parser)
			return -EBUSY;

	/* Insert new parser at front of list, so that a user-loaded module can
	 * (in case of some misdetection) override other parsers */
	parser->next = partition_parsers;
	partition_parsers = parser;
	return 0;
}

#ifdef CONFIG_MODULES
int unregister_partbl( struct partition_parser *parser )
{
	struct partition_parser **p;
    
	for( p = &partition_parsers; *p; p = &(*p)->next ) {
		if (*p == parser) {
			*p = parser->next;
			parser->next = NULL;
			return 0;
		}
	}
	return -EINVAL;
}
#endif

#ifdef CONFIG_PROC_FS
static struct proc_dir_entry *proc_partbl;

static int partbl_read_proc(char *buf, char **start, off_t offset,
			    int len, int *eof, void *private)
{
	struct partition_parser *p;

	len = 0;
	for( p = partition_parsers; p && len < 4000; p = p->next )
		len += sprintf(buf + len, "%s\n", p->name );
	*start = buf + offset;
	return len > offset ? len - offset : 0;
}
#endif

__initfunc(static void partbl_init( void ))
{
	extern void msdos_partbl_init(void);
	extern void osf_partbl_init(void);
	extern void sun_partbl_init(void);
	extern void amiga_partbl_init(void);
	extern void atari_partbl_init(void);
	extern void mac_partbl_init(void);
    
	if (HAVE_MAC_PARTITION)
		mac_partbl_init();
	if (HAVE_ATARI_PARTITION)
		atari_partbl_init();
	if (HAVE_AMIGA_PARTITION)
		amiga_partbl_init();
	if (HAVE_SUN_PARTITION)
		sun_partbl_init();
	if (HAVE_OSF_PARTITION)
		osf_partbl_init();
	if (HAVE_MSDOS_PARTITION)
		msdos_partbl_init();

#ifdef CONFIG_PROC_FS
	proc_partbl = create_proc_entry("partition-formats", 0, 0);
	if (proc_partbl)
		proc_partbl->read_proc = partbl_read_proc;
#endif
}

/* This function is used to re-read partition tables for removable disks.
   Much of the cleanup from the old partition tables should have already been
   done */

/* This function will re-read the partition tables for a given device,
and set things back up again.  There are some important caveats,
however.  You must ensure that no one is using the device, and no one
can start using the device while this function is being executed. */

void resetup_one_dev(struct gendisk *dev, int drive)
{
	int i;
	int first_minor	= drive << dev->minor_shift;
	int end_minor	= first_minor + dev->max_p;

	blk_size[dev->major] = NULL;
	current_minor = 1 + first_minor;
	check_partition(dev, MKDEV(dev->major, first_minor));

 	/*
 	 * We need to set the sizes array before we will be able to access
 	 * any of the partitions on this device.
 	 */
	if (dev->sizes != NULL) {	/* optional safeguard in ll_rw_blk.c */
		for (i = first_minor; i < end_minor; i++)
			dev->sizes[i] = dev->part[i].nr_sects >> (BLOCK_SIZE_BITS - 9);
		blk_size[dev->major] = dev->sizes;
	}
}

static inline void setup_dev(struct gendisk *dev)
{
	int i, drive;
	int end_minor	= dev->max_nr * dev->max_p;

	blk_size[dev->major] = NULL;
	for (i = 0 ; i < end_minor; i++) {
		dev->part[i].start_sect = 0;
		dev->part[i].nr_sects = 0;
	}
	dev->init(dev);	
	for (drive = 0 ; drive < dev->nr_real ; drive++) {
		int first_minor	= drive << dev->minor_shift;
		current_minor = 1 + first_minor;
		check_partition(dev, MKDEV(dev->major, first_minor));
	}
	if (dev->sizes != NULL) {	/* optional safeguard in ll_rw_blk.c */
		for (i = 0; i < end_minor; i++)
			dev->sizes[i] = dev->part[i].nr_sects >> (BLOCK_SIZE_BITS - 9);
		blk_size[dev->major] = dev->sizes;
	}
}

__initfunc(void device_setup(void))
{
	extern void console_map_init(void);
#ifdef CONFIG_PARPORT
	extern int parport_init(void);
#endif
#ifdef CONFIG_MD_BOOT
        extern void md_setup_drive(void) __init;
#endif
#ifdef CONFIG_FC4_SOC
	extern int soc_probe(void);
#endif
	struct gendisk *p;

#ifdef CONFIG_PARPORT
	parport_init();
#endif
	chr_dev_init();
	blk_dev_init();
	sti();
#ifdef CONFIG_FC4_SOC
	/* This has to be done before scsi_dev_init */
	soc_probe();
#endif
#ifdef CONFIG_SCSI
	scsi_dev_init();
#endif
#ifdef CONFIG_INET
	net_dev_init();
#endif
#ifdef CONFIG_VT
	console_map_init();
#endif

	partbl_init();
	for (p = gendisk_head ; p ; p=p->next)
		setup_dev(p);

#ifdef CONFIG_BLK_DEV_RAM
#ifdef CONFIG_BLK_DEV_INITRD
	if (initrd_start && mount_initrd) initrd_load();
	else
#endif
	rd_load();
#endif
#ifdef CONFIG_MD_BOOT
        md_setup_drive();
#endif
}

#ifdef CONFIG_PROC_FS
int get_partition_list(char * page)
{
	struct gendisk *p;
	char buf[8];
	int n, len;

	len = sprintf(page, "major minor  #blocks  name\n\n");
	for (p = gendisk_head; p; p = p->next) {
		for (n=0; n < (p->nr_real << p->minor_shift); n++) {
			if (p->part[n].nr_sects && len < PAGE_SIZE - 80) {
				len += sprintf(page+len,
					       "%4d  %4d %10d %s\n",
					       p->major, n, p->sizes[n],
					       disk_name(p, n, buf));
			}
		}
	}
	return len;
}
#endif

/* symbols visible for partbl_* modules */
EXPORT_SYMBOL(add_partition);
EXPORT_SYMBOL(get_ptable_blocksize);
EXPORT_SYMBOL(current_minor);
EXPORT_SYMBOL(register_partbl);
EXPORT_SYMBOL(unregister_partbl);
