/*
 *  partbl_atari.c -- Atari partition table parser
 *
 *  Code extracted from
 *  linux/drivers/block/genhd.c
 *
 */

#include <linux/module.h>

#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/ctype.h>
#include <linux/init.h>
#include <linux/atari_rootsec.h>
#include <asm/byteorder.h>


/* ++guenther: this should be settable by the user ("make config")?.
 * roman: doesn't use much code, better always include it :-)
 */
#define ICD_PARTS

/* check if a partition entry looks valid -- Atari format is assumed if at
 * least one of the primary entries is ok this way */
#define	VALID_PARTITION(pi,hdsiz)					     \
    (((pi)->flg & 1) &&							     \
     isalnum((pi)->id[0]) && isalnum((pi)->id[1]) && isalnum((pi)->id[2]) && \
     be32_to_cpu((pi)->st) <= (hdsiz) &&				     \
     be32_to_cpu((pi)->st) + be32_to_cpu((pi)->siz) <= (hdsiz))

static int atari_partition (struct gendisk *hd, kdev_t dev,
			    unsigned long first_sector)
{
  int minor = current_minor, m_lim = current_minor + hd->max_p;
  int i;
  struct buffer_head *bh;
  struct rootsector *rs;
  struct partition_info *pi;
  u32 extensect;
  u32 hd_size;
#ifdef ICD_PARTS
  int part_fmt = 0; /* 0:unknown, 1:AHDI, 2:ICD/Supra */
#endif

  MOD_INC_USE_COUNT;
  bh = bread (dev, 0, get_ptable_blocksize(dev));
  if (!bh)
    {
      printk (" unable to read block 0\n");
      MOD_DEC_USE_COUNT;
      return -1;
    }

  rs = (struct rootsector *) bh->b_data;
  hd_size = hd->part[current_minor-1].nr_sects;
  if (!VALID_PARTITION( &rs->part[0], hd_size ) &&
      !VALID_PARTITION( &rs->part[1], hd_size ) &&
      !VALID_PARTITION( &rs->part[2], hd_size ) &&
      !VALID_PARTITION( &rs->part[3], hd_size )) {
      /* if there's no valid primary partition, assume that no Atari format
       * partition table (there's no reliable magic or the like :-() */
      MOD_DEC_USE_COUNT;
      return 0;
  }

  printk (" AHDI");
  pi = &rs->part[0];
  for (; pi < &rs->part[4] && minor < m_lim; minor++, pi++)
    {
      if (pi->flg & 1)
	/* active partition */
	{
	  if (memcmp (pi->id, "XGM", 3) == 0)
	    /* extension partition */
	    {
	      struct rootsector *xrs;
	      struct buffer_head *xbh;
	      u32 partsect;

#ifdef ICD_PARTS
	      part_fmt = 1;
#endif
	      printk(" XGM<");
	      partsect = extensect = be32_to_cpu(pi->st);
	      while (1)
		{
		  xbh = bread (dev, partsect / 2, 1024);
		  if (!xbh)
		    {
		      printk (" block %u read failed\n", partsect);
		      brelse(bh);
		      MOD_DEC_USE_COUNT;
		      return -1;
		    }
		  if (partsect & 1)
		    xrs = (struct rootsector *) &xbh->b_data[512];
		  else
		    xrs = (struct rootsector *) &xbh->b_data[0];

		  /* ++roman: valid bit must be set in one of the first 3
		   * slots */
		  for( i = 0; i < 3; ++i )
		      if (xrs->part[0].flg & 1)
			  break;
		  if (i == 3) {
		    printk( "\nNo valid subpartition in extended "
			    "partition!\n" );
		    break;
		  }
		  
		  add_partition(hd, minor,
				partsect + be32_to_cpu(xrs->part[i].st),
				be32_to_cpu(xrs->part[i].siz));

		  /* the slot following must be either invalid (end of list)
		   * or another XGM entry */
		  if (!(xrs->part[i+1].flg & 1)) {
		    brelse( xbh );
		    break;
		  }
		  if (memcmp( xrs->part[i+1].id, "XGM", 3 ) != 0) {
		    printk( "\nLink ID in extended partition is not XGM!\n" );
		    brelse( xbh );
		    break;
		  }

		  partsect = be32_to_cpu(xrs->part[i+1].st) + extensect;
		  brelse (xbh);
		  minor++;
		  if (minor >= m_lim) {
		    printk( "\nMaximum number of partitions reached!\n" );
		    break;
		  }
		}
	      printk(" >");
	    }
	  else
	    {
	      /* we don't care about other id's */
	      add_partition (hd, minor, be32_to_cpu(pi->st),
			     be32_to_cpu(pi->siz));
	    }
	}
    }
#ifdef ICD_PARTS
  if ( part_fmt!=1 ) /* no extended partitions -> test ICD-format */
  {
    pi = &rs->icdpart[0];
    /* sanity check: no ICD format if first partition invalid */
    if (memcmp (pi->id, "GEM", 3) == 0 ||
        memcmp (pi->id, "BGM", 3) == 0 ||
        memcmp (pi->id, "LNX", 3) == 0 ||
        memcmp (pi->id, "SWP", 3) == 0 ||
        memcmp (pi->id, "RAW", 3) == 0 )
    {
      printk(" ICD<");
      for (; pi < &rs->icdpart[8] && minor < m_lim; minor++, pi++)
      {
        /* accept only GEM,BGM,RAW,LNX,SWP partitions */
        if (pi->flg & 1 && 
            (memcmp (pi->id, "GEM", 3) == 0 ||
             memcmp (pi->id, "BGM", 3) == 0 ||
             memcmp (pi->id, "LNX", 3) == 0 ||
             memcmp (pi->id, "SWP", 3) == 0 ||
             memcmp (pi->id, "RAW", 3) == 0) )
        {
          part_fmt = 2;
	  add_partition (hd, minor, be32_to_cpu(pi->st), be32_to_cpu(pi->siz));
        }
      }
      printk(" >");
    }
  }
#endif
  brelse (bh);
  printk ("\n");
  MOD_DEC_USE_COUNT;
  return 1;
}

static struct partition_parser atari_partition_parser =
    { NULL, "atari", atari_partition };

#ifndef MODULE
__initfunc(void atari_partbl_init( void ))
{
	(void)register_partbl( &atari_partition_parser );
}
#else
int init_module(void)
{
	return register_partbl( &atari_partition_parser );
}

void cleanup_module(void)
{
	(void)unregister_partbl( &atari_partition_parser );
}
#endif
