/*
 * lp_ioext: Parallel printer support for GVP IO-Extender.
 *
 * Author:	Steve Bennett, <msteveb@ozemail.com.au>
 *
 */
#define __NO_VERSION__

#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/lp_m68k.h>
#include <linux/16c552.h>
#include <asm/irq.h>
#include <asm/setup.h>

#include "ioext.h"

#define DEBUG 0

/* Match up par device numbers with our control structure */
#define PAR_DEV(DEV)  ((IOExtInfoType *)lp_table[DEV]->base)

static void lp_out(int c, int dev);
static int lp_busy(int dev);
static int lp_pout(int dev);
static int lp_online(int dev);
static int lp_open(int dev);
static void lp_release(int dev);

/*
 * Check for and handle an interrupt for this PARALLEL device.
 *
 */
void ioext_lp_interrupt(int dev, int *spurious_count)
{
	/* Acknowledge the interrupt */
	(void)PAR_DEV(dev)->board->par.STATUS;

	if (!(lp_table[dev]->flags & LP_BUSY)) {
		/*printk("OOPS: lp_interrupt called, but lp not in use\n");*/
		return;
	}

	/* This is cleared in lp_out() */
	++*spurious_count;

	if (!(PAR_DEV(dev)->board->par.STATUS & LP_PBUSY)) {
#if DEBUG
		printk("ioext_lp_interrupt(): busy\n");
#endif
		/* Can't have been an interrupt for us if we are still busy */
		return;
	}

	lp_interrupt(dev);
}

static void lp_out(int c, int dev)
{
	int wait;
	IOEXT_struct *board = PAR_DEV(dev)->board;

#if DEBUG
	printk("lp_out(c=%02X, dev=%d)\n", c, dev);
#endif

	PAR_DEV(dev)->spurious_count = 0;

	board->par.DATA = c;

	/* must wait before taking strobe high, and after taking strobe
	   low, according to the spec.  Some printers need it, others don't. */
	wait = lp_table[dev]->wait;
	while (wait--) {
	}
	/* control port takes strobe high */
	board->par.CTRL |= LP_PSTROBE;

	wait = lp_table[dev]->wait;
	while (wait) {
		wait--;
	}

	/* take strobe low */
	board->par.CTRL &= ~LP_PSTROBE;
}

static int lp_busy(int dev)
{
	return(!(PAR_DEV(dev)->board->par.STATUS & LP_PBUSY));
}

static int lp_pout(int dev)
{
	return(PAR_DEV(dev)->board->par.STATUS & LP_POUTPA);
}

static int lp_online(int dev)
{
	return(PAR_DEV(dev)->board->par.STATUS & LP_PSELECD);
}

static int lp_open(int dev)
{
#if defined(CONFIG_GVPIOEXT_PLIP) || defined(CONFIG_GVPIOEXT_PLIP_MODULE)
	/* Yes, there is a race condition here. Fix it later */
	if (PAR_DEV(dev)->par_use & IOEXT_PAR_PLIP) {
		/* Can't open if PLIP is in use */
#if DEBUG
		printk("par is in use by plip\n");
#endif
		return(-EBUSY);
	}
#endif
	MOD_INC_USE_COUNT;
	PAR_DEV(dev)->par_use |= IOEXT_PAR_LP;
	return(0);
}

static void lp_release(int dev)
{
	PAR_DEV(dev)->par_use &= ~IOEXT_PAR_LP;
	MOD_DEC_USE_COUNT;
}

/* Only support one printer port for now.
*/
struct lp_struct ioext_lp_table[] = {
  {
	  "IO-extender parallel port",	/* name */
	  0,				/* irq */
	  lp_out,
	  lp_busy,
	  lp_pout,
	  lp_online,
	  0,
	  NULL,				/* ioctl */
	  lp_open,
	  lp_release,
	  LP_EXIST,
	  LP_INIT_CHAR,
	  LP_INIT_TIME,
	  LP_INIT_WAIT,
	  NULL,
	  NULL,
	  LP_IOEXT
  },
};
